import abc
import functools

from . import _version
from . import _six

name = 'doorbell'
__version__ = _version.get_versions()['version']


class Visitee(_six.with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def accept(self, visitor):
        """accept a `Visitor`

        Call the appropriate `Visitor.visit_*` method and return the result.

        Args:
            visitor (Visitor): the visitor to accept Returns:
            any value returned by the visitor

        """
        raise NotImplementedError()


class _VisitorMethod(object):
    """a method that is used as a visitor

    Internal type to mark a method as a visitor method.  By default, all
    methods named "visit_*" will be marked as a visitor method.  If `skip` is
    True, this method will *not* be wrapped as a visitor method.

    Attributes:
        function (Callable): the visitor method
        skip (bool): if True, this method is not used as a visitor

    Args:
        function (Callable): the visitor method
        skip (bool): default False

    """
    def __init__(self, function, skip=False):
        self.function = function
        # skip is currently not used, but I can't resist
        self.skip = skip


class _MetaVisitor(type):
    def __init__(cls, name, bases, attrs):
        for k, v in attrs.items():
            if isinstance(v, _VisitorMethod):
                v = v.function
            elif k.startswith('visit_'):
                v = cls.visitor_method(v).function
            attrs[k] = v
            setattr(cls, k, v)
        super(_MetaVisitor, cls).__init__(name, bases, attrs)


class Visitor(_six.with_metaclass(_MetaVisitor, object)):
    """A basic visitor.

    This implementation does nothing special, but provides groundwork for
    child classes.

    Attributes:
        _visiting (bool): whether this is currently visiting a visitee

    """
    def __init__(self, *args, **kwargs):
        super(Visitor, self).__init__(*args, **kwargs)
        self._visiting = False

    @classmethod
    def visitor_method(cls, func):
        """Wrapper to mark non-default method as a visitor method.

        >>> class MyClass(Visitor):
        ...    @Visitor.visitor_method
        ...    def unusual_method_name(self, subject):
        ...        pass

        """
        # wrap func with _visit_wrapper
        wrapper = _six.partialmethod(cls._visit_wrapper, function=func)
        wrapper = functools.wraps(func)(wrapper)
        # mark as a visitor method
        return _VisitorMethod(wrapper)

    @classmethod
    def non_visitor_method(cls, func):
        """Wrapper to mark method as *not* a visitor method.

        >>> class MyClass(Visitor):
        ...    @Visitor.non_visitor_method
        ...    def visit_DontUseThisMethod(self, subject):
        ...        pass

        """
        return _VisitorMethod(func, skip=True)

    def _visit_wrapper(self, *args, **kwargs):
        """Wrapper for visitor methods.

        This method wraps all visitor methods.  This is intended to be
        overridden, and need not be called when overridden.  The wrapped
        method is always passed in as a keyword argument `function`.  This
        must pass all other positional and keyword arguments to that function.

        Args:
            function: the function to wrap; always passed as keyword argument

        """
        return kwargs.pop('function')(self, *args, **kwargs)


class WrappingVisitor(Visitor):
    """A visitor class that wraps each call with pre and post methods.

    """
    def _visit_wrapper(self, *args, **kwargs):
        func = kwargs['function']
        if self._visiting:
            pre = self._wrap_each_pre
            post = self._wrap_each_post
            return post(func(self, *pre(*args)))
        else:
            self._visiting = True
            pre = self._wrap_all_pre
            post = self._wrap_all_post
            try:
                return post(self._visit_wrapper(*pre(*args), **kwargs))
            finally:
                self._visiting = False

    def _wrap_each_pre(self, *args):
        """Method called before each visit.

        Returns:
            all arguments to be passed to visitor method

        """
        return args

    def _wrap_each_post(self, arg):
        """Method called after each visit.

        Returns:
            return value from each visitor method

        """
        return arg

    def _wrap_all_pre(self, *args):
        """Method called before any visit method is called.

        This is the first method called when a visitor method is called.

        Returns:
            arguments passed to :func:`_wrap_each_pre`

        """
        return args

    def _wrap_all_post(self, arg):
        """Method called after all visit methods are called.

        This is the last method called when a visitor method is called.

        Returns:
            final return value from a visitor method call

        """
        return arg


class CascadingVisitor(WrappingVisitor):
    """Visits children first.

    This visitor visits all child nodes of a visitee first, cascading through
    all children.

    Visitor methods are passed the visitee, then a list of return values from
    visiting the children, then all remaining arguments.

    """
    def _gather_children(self, subject):
        """Gather children from a visitee.

        Default implementation simply returns `subject.children`.

        """
        return subject.children

    def _wrap_each_pre(self, subject, *args):
        children = [c.accept(self) for c in self._gather_children(subject)]
        args = list(args)
        args.insert(0, children)
        args.insert(0, subject)
        return args



import doorbell
import functools
import operator


class Value(doorbell.Visitee):
    def __init__(self, value=0):
        self.value = value
        self.children = []

    def accept(self, visitor):
        return visitor.visit_Value(self)


class Add(Value):
    def accept(self, visitor):
        return visitor.visit_Add(self)


class Mult(Add):
    def accept(self, visitor):
        return visitor.visit_Mult(self)


class ManyArgs(Value):
    def accept(self, visitor, *args):
        return visitor.visit_ManyArgs(self, *args)


class Visitor(doorbell.CascadingVisitor):
    def visit_Value(self, obj, children):
        return obj.value

    def visit_Add(self, obj, children):
        return functools.reduce(operator.add, children, 0)

    def visit_Mult(self, obj, children):
        return functools.reduce(operator.mul, children, 1)

    def visit_ManyArgs(self, obj, children, *args):
        return len(args)


class TestVisitee:
    def test_add(self):
        one = Value(1)
        add = Add()
        add.children.append(one)
        add.children.append(one)
        v = Visitor()
        r = add.accept(v)
        assert r == 2

    def test_multiply(self):
        one = Value(1)
        mult = Mult()
        mult.children.append(one)
        mult.children.append(one)
        v = Visitor()
        r = mult.accept(v)
        assert r == 1

    def test_combined(self):
        one = Value(1)
        add = Add()
        add.children.extend((one, one))
        mult = Mult()
        mult.children.extend((add, add))
        v = Visitor()
        r = mult.accept(v)
        assert r == 4

    def test_many_arguments(self):
        """visitor method should receive all additional arguments"""
        m = ManyArgs()
        v = Visitor()
        assert m.accept(v, 1, 2, 3) == 3



# various 2to3 fixes
import functools

class partialmethod(object):
    def __init__(self, wrapped_function, *args, **kwargs):
        self.function = wrapped_function
        self.args = args
        self.kwargs = kwargs
    def __get__(self, instance, *args, **kwargs):
        if instance is None:
            return self
        else:
            return functools.partial(self.function, instance, *self.args, **self.kwargs)

# borrowed from six.py:
# Copyright (c) 2010-2018 Benjamin Peterson
# https://github.com/benjaminp/six
# commit: a611f60
def with_metaclass(meta, *bases):
    class metaclass(type):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
        @classmethod
        def __prepare__(cls, name, this_bases):
            return meta.__prepare__(name, bases)
    return type.__new__(metaclass, 'temporary_class', (), {})



import doorbell


class TestNonVisitorDecorator:
    class Value(doorbell.Visitee):
        def __init__(self, value=0):
            self.value = value

        def accept(self, visitor):
            return visitor.visit_Value(self)

    class Visitor(doorbell.WrappingVisitor):
        def _wrap_each_pre(*args, **kwargs):
            raise NotImplementedError()

        @doorbell.Visitor.non_visitor_method
        def visit_Value(self, obj):
            return obj.value

    def test_visitor_method(self):
        v = self.Value(1)
        a = self.Visitor()
        assert v.accept(a) == 1


class TestBasicWrapping:
    class Value(doorbell.Visitee):
        def __init__(self, value=0):
            self.value = value

        def accept(self, visitor):
            return visitor.visit_Value(self)

    class Visitor(doorbell.WrappingVisitor):
        def visit_Value(self, obj):
            return obj.value

    def test_visitor_method(self):
        v = self.Value(1)
        a = self.Visitor()
        assert v.accept(a) == 1



import doorbell


class TestVisitorDecorator:
    class Value(doorbell.Visitee):
        def __init__(self, value=0):
            self.value = value

        def accept(self, visitor):
            return visitor.goto_Value(self)

    class Visitor(doorbell.Visitor):
        @doorbell.Visitor.visitor_method
        def goto_Value(self, obj):
            return obj.value

    def test_visitor_method(self):
        v = self.Value(1)
        a = self.Visitor()
        assert v.accept(a) == 1



from setuptools import find_packages
from setuptools import setup

import versioneer

test_deps = [
    'coverage',
    'pytest',
    'pytest-cov',
    'sphinx_rtd_theme',
    'tox',
    'versioneer',
    ],
extras = {
    'test': test_deps,
}

setup(
    name="doorbell",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Tim Hartman",
    author_email="tbhartman@gmail.com",
    description=("A visitor pattern implementation for Python"),
    license="MIT",
    keywords="visitor",
    url="https://github.com/tbhartman/doorbell",
    packages=find_packages('src', exclude=['test*']),
    package_dir={'': 'src'},
    install_requires=[],
    tests_require=test_deps,
    extras_require=extras,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
